import getDebugger from '../util/debug.es6';
import * as tables from './tables.es6';
import {PlayBuffer, RenderBuffer} from '../util/player.es6';
import UInt8 from '../types/UInt8.es6';

const PHONEME_PERIOD = 1;
const PHONEME_QUESTION = 2;
const RISING_INFLECTION = 1;
const FALLING_INFLECTION = 255;

const FLAG_PLOSIVE  = 0x0001;
const FLAG_STOPCONS = 0x0002; /* stop consonant */
const FLAG_VOICED   = 0x0004;
 /* 0x08 */
const FLAG_DIPTHONG = 0x0010;
const FLAG_DIP_YX   = 0x0020; /* dipthong ending with YX */
const FLAG_CONSONANT= 0x0040;
const FLAG_VOWEL    = 0x0080;
const FLAG_PUNCT    = 0x0100;
 /* 0x200 */
const FLAG_ALVEOLAR = 0x0400;
const FLAG_NASAL    = 0x0800;
const FLAG_LIQUIC   = 0x1000;  /* liquic consonant */
const FLAG_FRICATIVE= 0x2000;


const pR    = 23;
const pD    = 57;
const pT    = 69;
const BREAK = 254;
const END   = 255;

export default class Renderer {
  /**
   *
   * @param {object} options
   * @param {Boolean} options.phonetic Default false.
   * @param {Boolean} options.singmode Default false.
   * @param {Boolean} options.debug Default false.
   * @param {Number} options.pitch Default 64.
   * @param {Number} options.speed Default 72.
   * @param {Number} options.mouth Default 128.
   * @param {Number} options.throat Default 128.
   *
   * @returns {Function}
   * @constructor
   */
  constructor(options) {
    // Default values.
    const opts = {bufferlength: 10};
    opts.speed = (options.speed || 72) & 0xFF;
    opts.pitch = (options.pitch || 64) & 0xFF;
    opts.mouth = (options.mouth || 128) & 0xFF;
    opts.throat = (options.throat || 128) & 0xFF;
    opts.singmode = options.singmode || false;
    opts.debug = options.debug || false;

    this.getOptions = () => opts;
    this.debug = getDebugger(opts.debug);

    this.pitches    = new Uint8Array(256);
    this.frequency1 = new Uint8Array(256);
    this.frequency2 = new Uint8Array(256);
    this.frequency3 = new Uint8Array(256);
    this.amplitude1 = new Uint8Array(256);
    this.amplitude2 = new Uint8Array(256);
    this.amplitude3 = new Uint8Array(256);
    this.sampledConsonantFlag = new Uint8Array(256);

    this.stress = new Uint8Array(256); //numbers from 0 to 8
    this.phonemeLength = new Uint8Array(256); //tab40160
    this.phonemeindex = new Uint8Array(256);
    this.phonemeIndexOutput = new Uint8Array(60); //tab47296
    this.stressOutput = new Uint8Array(60); //tab47365
    this.phonemeLengthOutput = new Uint8Array(60); //tab47416
  }

  speak (_input) {
    const input = new Uint8Array(256);
    let l = _input.length;
    if (l > 254) l = 254;
    for(let i=0; i<l; i++)
      input[i] = _input[i];
    input[l] = 0;

    this.SAMMain(input);

    const bufferLength = Math.floor(this.bufferpos / 50);
    // Now push buffer to wave player.
    const audiobuffer = new Float32Array(bufferLength);

    for(let i=0; i < bufferLength; i++) {
      // audiobuffer[i] = ((this.buffer & 0xFF) - 128) / 256;
      audiobuffer[i] = (this.buffer[i] - 128) / 256;
    }
    this.debug.log(audiobuffer.length);

    PlayBuffer(audiobuffer);

    const wavebuffer = this.buffer.slice(0, bufferLength);
    // RenderBuffer(wavebuffer);
  }

  get pitch () { return this.getOptions().pitch }
  get singmode () { return this.getOptions().singmode }

  get bufferpos () { return this._bufferpos }
  set bufferpos (v) {
    this._bufferpos = v & 0xFFFFFFFF;
    if (this._bufferpos / 50 > this.buffer.length) {
      throw new Error('Buffer overflow!');
    }
  }

  ///////////////////////////////////////////////////////////////// sam.c from here on.

  /**
   * Prepare the wave form buffer.
   *
   * @param {Uint8Array} input The input data.
   *
   * @returns {number} 1 on success, 0 on failure.
   * @constructor
   */
  SAMMain(input) {
    this.Init();
    /* FIXME: At odds with assignment in Init() */
    this.phonemeindex[255] = 32; //to prevent buffer overflow

     if (!this.Parser1(input)) return 0;
    this.PrintPhonemes(this.phonemeindex, this.phonemeLength, this.stress);
    this.Parser2();
    this.CopyStress();
    this.SetPhonemeLength();
    this.AdjustLengths();
    this.Code41240();
    let X = new UInt8();
    do {
      if (this.phonemeindex[X.get()] > 80) {
        this.phonemeindex[X.get()] = END;
        break; // error: delete all behind it
      }
    } while (X.inc().get() !== 0);
    this.InsertBreath();

    this.PrintPhonemes(this.phonemeindex, this.phonemeLength, this.stress);

    this.PrepareOutput();
    return 1;
  }

  Init() {
    this.freq1data = tables.freq1data.slice();
    this.freq2data = tables.freq2data.slice();
    this.SetMouthThroat(this.getOptions().mouth, this.getOptions().throat);

    // TODO, check for free the memory, 10 seconds of output should be more than enough
    this.buffer = new Uint8Array(22050*(this.getOptions().bufferlength || 10));
    this.bufferpos = 0;

    for(let i=0; i<256; i++) {
      this.stress[i] = 0;
      this.phonemeLength[i] = 0;
    }

    for(let i=0; i<60; i++) {
      this.phonemeIndexOutput[i] = 0;
      this.stressOutput[i] = 0;
      this.phonemeLengthOutput[i] = 0;
    }
    //to prevent buffer overflow
    // ML : changed from 32 to 255 to stop freezing with long inputs
    this.phonemeindex[255] = END;
  }

  /**
   SAM's voice can be altered by changing the frequencies of the
   mouth formant (F1) and the throat formant (F2). Only the voiced
   phonemes (5-29 and 48-53) are altered.
   */
  SetMouthThroat(mouth, throat) {
    const trans = (a, b) => {
      return ((((a & 0xFF) * (b & 0xFF)) >> 8) & 0xFF) << 1;
    };

    let initialFrequency;
    let newFrequency = 0;

    // mouth formants (F1) 5..29
    let mouthFormants5_29 = [
      0, 0, 0, 0, 0, 10,
      14, 19, 24, 27, 23, 21, 16, 20, 14, 18, 14, 18, 18,
      16, 13, 15, 11, 18, 14, 11, 9, 6, 6, 6
    ];

    // throat formants (F2) 5..29
    let throatFormants5_29 = [
      255, 255,
      255, 255, 255, 84, 73, 67, 63, 40, 44, 31, 37, 45, 73, 49,
      36, 30, 51, 37, 29, 69, 24, 50, 30, 24, 83, 46, 54, 86,
    ];

    // there must be no zeros in this 2 tables
    // formant 1 frequencies (mouth) 48..53
    let mouthFormants48_53 = [19, 27, 21, 27, 18, 13];

    // formant 2 frequencies (throat) 48..53
    let throatFormants48_53 = [72, 39, 31, 43, 30, 34];

    let pos = 5;

    // recalculate formant frequencies 5..29 for the mouth (F1) and throat (F2)
    while(pos < 30) {
      // recalculate mouth frequency
      initialFrequency = mouthFormants5_29[pos];
      if (initialFrequency !== 0) newFrequency = trans(mouth, initialFrequency);
      // FIXME: overrides table!
      this.freq1data[pos] = newFrequency;

      // recalculate throat frequency
      initialFrequency = throatFormants5_29[pos];
      if(initialFrequency !== 0) newFrequency = trans(throat, initialFrequency);
      // FIXME: overrides table!
      this.freq2data[pos] = newFrequency;
      pos++;
    }

    // recalculate formant frequencies 48..53
    pos = 0;
    while(pos < 6) {
      // recalculate F1 (mouth formant)
      initialFrequency = mouthFormants48_53[pos];
      newFrequency = trans(mouth, initialFrequency);
      // FIXME: overrides table!
      this.freq1data[pos+48] = newFrequency;

      // recalculate F2 (throat formant)
      initialFrequency = throatFormants48_53[pos];
      newFrequency = trans(throat, initialFrequency);
      // FIXME: overrides table!
      this.freq2data[pos+48] = newFrequency;
      pos++;
    }
  }


  /**
   * The input[] buffer contains a string of phonemes and stress markers along
   * the lines of:
   *
   *     DHAX KAET IHZ AH5GLIY. <0x9B>
   *
   * The byte 0x9B marks the end of the buffer. Some phonemes are 2 bytes
   * long, such as "DH" and "AX". Others are 1 byte long, such as "T" and "Z".
   * There are also stress markers, such as "5" and ".".
   *
   * The first character of the phonemes are stored in the table signInputTable1[].
   * The second character of the phonemes are stored in the table signInputTable2[].
   * The stress characters are arranged in low to high stress order in stressInputTable[].
   *
   * The following process is used to parse the input[] buffer:
   *
   * Repeat until the <0x9B> character is reached:
   *
   *        First, a search is made for a 2 character match for phonemes that do not
   *        end with the '*' (wildcard) character. On a match, the index of the phoneme
   *        is added to phonemeIndex[] and the buffer position is advanced 2 bytes.
   *
   *        If this fails, a search is made for a 1 character match against all
   *        phoneme names ending with a '*' (wildcard). If this succeeds, the
   *        phoneme is added to phonemeIndex[] and the buffer position is advanced
   *        1 byte.
   *
   *        If this fails, search for a 1 character match in the stressInputTable[].
   *        If this succeeds, the stress value is placed in the last stress[] table
   *        at the same index of the last added phoneme, and the buffer position is
   *        advanced by 1 byte.
   *
   *        If this fails, return a 0.
   *
   * On success:
   *
   *    1. phonemeIndex[] will contain the index of all the phonemes.
   *    2. The last index in phonemeIndex[] will be 255.
   *    3. stress[] will contain the stress value for each phoneme
   *
   * input[] holds the string of phonemes, each two bytes wide
   * signInputTable1[] holds the first character of each phoneme
   * signInputTable2[] holds te second character of each phoneme
   * phonemeIndex[] holds the indexes of the phonemes after parsing input[]
   *
   * The parser scans through the input[], finding the names of the phonemes
   * by searching signInputTable1[] and signInputTable2[]. On a match, it
   * copies the index of the phoneme into the phonemeIndexTable[].
   *
   * The character <0x9B> marks the end of text in input[]. When it is reached,
   * the index 255 is placed at the end of the phonemeIndexTable[], and the
   * function returns with a 1 indicating success.
   *
   * @param {Uint8Array} input The input values.
   */
  Parser1(input) {
    const full_match = (sign1, sign2) => {
      let Y = 0;
      do {
        // GET FIRST CHARACTER AT POSITION Y IN signInputTable
        // --> should change name to PhonemeNameTable1
        let A = tables.signInputTable1[Y];

        if (A === sign1) {
          A = tables.signInputTable2[Y];
          // NOT A SPECIAL AND MATCHES SECOND CHARACTER?
          if ((A !== 42 /* '*' */) && (A === sign2)) {
            return Y;
          }
        }
      } while (++Y !== 81);
      return -1;
    };

    const wild_match = (sign1) => {
      let Y = 0;
      do {
        if (tables.signInputTable2[Y] === 42 /* '*' */) {
          if (tables.signInputTable1[Y] === sign1) {
            return Y;
          }
        }
      } while (++Y !== 81);
      return -1;
    };

    let i;
    let sign1;
    let sign2;
    let position = 0;
    let srcpos   = 0;

    // Clear the stress table.
    for(i=0; i<256; i++) {
      this.stress[i] = 0;
    }

    while((sign1 = input[srcpos]) !== 155) { // 155 (\233) is end of line marker
      sign2 = input[++srcpos];
      let match = 0;
      if ((match = full_match(sign1, sign2)) !== -1) {
        // Matched both characters (no wildcards)
        this.phonemeindex[position++] = match;
        ++srcpos; // Skip the second character of the input as we've matched it
      } else if ((match = wild_match(sign1)) !== -1) {
        // Matched just the first character (with second character matching '*'
        this.phonemeindex[position++] = match;
      } else {
        // Should be a stress character. Search through the
        // stress table backwards.
        match = 8; // End of stress table. FIXME: Don't hardcode.
        while ((sign1 !== tables.stressInputTable[match]) && (match>0)) --match;

        if (match === 0) {
          return 0; // failure
        }

        this.stress[position-1] = match; // Set stress for prior phoneme
      }
    } //while

    this.phonemeindex[position] = END;
    return 1;
  }

  /**
   * Insert a phoneme at the given position.
   *
   * @param {Number} position The position in the phoneme arrays to work set.
   * @param {Number} index    The phoneme index.
   * @param {Number} length   The phoneme length.
   * @param {Number} stress   The stress.
   * @constructor
   */
  Insert(position/*var57*/, index, length, stress) {
    // ML : always keep last safe-guarding 255
    for(let i = 253; i >= position; i--) {
      this.phonemeindex[i+1]  = this.phonemeindex[i];
      this.phonemeLength[i+1] = this.phonemeLength[i];
      this.stress[i+1]        = this.stress[i];
    }

    this.phonemeindex[position]  = index;
    this.phonemeLength[position] = length;
    this.stress[position]        = stress;
  }

  /**
   * Rewrites the phonemes using the following rules:
   *
   * <DIPTHONG ENDING WITH WX> -> <DIPTHONG ENDING WITH WX> WX
   * <DIPTHONG NOT ENDING WITH WX> -> <DIPTHONG NOT ENDING WITH WX> YX
   * UL -> AX L
   * UM -> AX M
   * <STRESSED VOWEL> <SILENCE> <STRESSED VOWEL> -> <STRESSED VOWEL> <SILENCE> Q <VOWEL>
   * T R -> CH R
   * D R -> J R
   * <VOWEL> R -> <VOWEL> RX
   * <VOWEL> L -> <VOWEL> LX
   * G S -> G Z
   * K <VOWEL OR DIPTHONG NOT ENDING WITH IY> -> KX <VOWEL OR DIPTHONG NOT ENDING WITH IY>
   * G <VOWEL OR DIPTHONG NOT ENDING WITH IY> -> GX <VOWEL OR DIPTHONG NOT ENDING WITH IY>
   * S P -> S B
   * S T -> S D
   * S K -> S G
   * S KX -> S GX
   * <ALVEOLAR> UW -> <ALVEOLAR> UX
   * CH -> CH CH' (CH requires two phonemes to represent it)
   * J -> J J' (J requires two phonemes to represent it)
   * <UNSTRESSED VOWEL> T <PAUSE> -> <UNSTRESSED VOWEL> DX <PAUSE>
   * <UNSTRESSED VOWEL> D <PAUSE>  -> <UNSTRESSED VOWEL> DX <PAUSE>
   */
  Parser2(phonemeindex, stress) {
    const rule_alveolar_uw = (X) => {
      // ALVEOLAR flag set?
      if ((tables.flags[this.phonemeindex[X-1]] & FLAG_ALVEOLAR) !== 0) {
        this.drule('<ALVEOLAR> UW -> <ALVEOLAR> UX');
        this.phonemeindex[X] = 16;
      }
    };

    const rule_ch = (X) => {
      this.drule('CH -> CH CH+1');
      this.Insert(X + 1, 43, 0, this.stress[X]);
    };

    const rule_j = (X) => {
      this.drule("J -> J J+1");
      this.Insert(X + 1, 45, 0, this.stress[X]);
    };

    const rule_g = (pos) => {
      // G <VOWEL OR DIPTHONG NOT ENDING WITH IY> -> GX <VOWEL OR DIPTHONG NOT ENDING WITH IY>
      // Example: GO

      let index = this.phonemeindex[pos+1];

      // If dipthong ending with YX, move continue processing next phoneme
      if ((index !== 255) && ((tables.flags[index] & FLAG_DIP_YX) === 0)) {
        // replace G with GX and continue processing next phoneme
        this.drule('G <VOWEL OR DIPTHONG NOT ENDING WITH IY> -> GX <VOWEL OR DIPTHONG NOT ENDING WITH IY>');
        this.phonemeindex[pos] = 63; // 'GX'
      }
    };

    const rule_dipthong = (p, pf, pos) => {
      // <DIPTHONG ENDING WITH WX> -> <DIPTHONG ENDING WITH WX> WX
      // <DIPTHONG NOT ENDING WITH WX> -> <DIPTHONG NOT ENDING WITH WX> YX
      // Example: OIL, COW

      // If ends with IY, use YX, else use WX
      const A = ((pf & FLAG_DIP_YX) !== 0) ? 21 : 20; // 'WX' = 20 'YX' = 21

      // Insert at WX or YX following, copying the stress
      if (A === 20) {this.drule('insert WX following dipthong NOT ending in IY sound');}
      if (A === 21) {this.drule('insert YX following dipthong ending in IY sound');}
      this.Insert((pos + 1) & 0xFF, A, 0, this.stress[pos]);

      if (p === 53 || p === 42 || p === 44) {
        if (p === 53) {
          // Example: NEW, DEW, SUE, ZOO, THOO, TOO
          rule_alveolar_uw(pos);
        } else if (p === 42) {
          // Example: CHEW
          rule_ch(pos, 0);
        } else if (p === 44) {
          // Example: JAY
          rule_j(pos, 0);
        }
      }
    };

    const change = (pos, val, ruleDescription) => {
      this.drule(ruleDescription);
      this.phonemeindex[pos] = val;
    };

    const ChangeRule = (position, rule, mem60, stress, descr) => {
      position = position & 0xFF;
      this.debug.log('RULE: %s', descr);
      this.phonemeindex[position] = rule;
      this.Insert(position + 1, mem60, 0, stress);
    };

    let pos = 0; //mem66;
    let p;

    while((p = this.phonemeindex[pos]) !== END) {
      this.debug.log('%d: %s', pos, String.fromCharCode(tables.signInputTable1[p], tables.signInputTable2[p]));

      if (p === 0) { // Is phoneme pause?
        ++pos;
        continue;
      }

      let pf = tables.flags[p];
      let prior = this.phonemeindex[pos-1];

      if ((pf & FLAG_DIPTHONG) !== 0) {
        rule_dipthong(p, pf, pos, 0);
      } else if (p === 78) {
        // Example: MEDDLE
        ChangeRule(pos, 13, 24, this.stress[pos], 'UL -> AX L');
      } else if (p === 79) {
        // Example: ASTRONOMY
        ChangeRule(pos, 13, 27, this.stress[pos], 'UM -> AX M');
      } else if (p === 80) {
        ChangeRule(pos, 13, 28, this.stress[pos], 'UN -> AX N');
      } // Example: FUNCTION
      else if ((pf & FLAG_VOWEL) && this.stress[pos]) {
        // RULE:
        //       <STRESSED VOWEL> <SILENCE> <STRESSED VOWEL> -> <STRESSED VOWEL> <SILENCE> Q <VOWEL>
        // EXAMPLE: AWAY EIGHT
        if (!this.phonemeindex[pos+1]) { // If following phoneme is a pause, get next
          p = this.phonemeindex[pos+2];
          if (p !== END && ((tables.flags[p] & FLAG_VOWEL) !== 0) && this.stress[pos+2]) {
            this.drule('Insert glottal stop between two stressed vowels with space between them');
            this.Insert(pos+2, 31, 0, 0); // 31 = 'Q'
          }
        }
      } else if (p === pR) { // RULES FOR PHONEMES BEFORE R
        if (prior === pT) {
          // Example: TRACK
          change(pos-1, 42, 'T R -> CH R');
        } else if (prior === pD) {
          // Example: DRY
          change(pos-1, 44, 'D R -> J R');
        } else if ((tables.flags[prior] & FLAG_VOWEL) !== 0) {
          // Example: ART
          change(pos, 18, '<VOWEL> R -> <VOWEL> RX');
        }
      } else if ((p === 24) && ((tables.flags[prior] & FLAG_VOWEL) !== 0)) {
        // Example: ALL
        change(pos, 19, '<VOWEL> L -> <VOWEL> LX');
      } else if (prior === 60 && p === 32) { // 'G' 'S'
        // Can't get to fire -
        //       1. The G -> GX rule intervenes
        //       2. Reciter already replaces GS -> GZ
        change(pos, 38, "G S -> G Z");
      } else if (p === 60) {
        rule_g(pos);
      } else {
        if (p === 72) {  // 'K'
          // K <VOWEL OR DIPTHONG NOT ENDING WITH IY> -> KX <VOWEL OR DIPTHONG NOT ENDING WITH IY>
          // Example: COW
          let Y = this.phonemeindex[pos+1];
          // If at end, replace current phoneme with KX
          if ((tables.flags[Y] & FLAG_DIP_YX) === 0 || Y === END) {
            // VOWELS AND DIPTHONGS ENDING WITH IY SOUND flag set?
            change(pos, 75, 'K <VOWEL OR DIPTHONG NOT ENDING WITH IY> -> KX <VOWEL OR DIPTHONG NOT ENDING WITH IY>');
            p  = 75;
            pf = tables.flags[p];
          }
        }

        // Replace with softer version?
        if ((tables.flags[p] & FLAG_PLOSIVE) && (prior === 32)) { // 'S'
          // RULE:
          //      S P -> S B
          //      S T -> S D
          //      S K -> S G
          //      S KX -> S GX
          // Examples: SPY, STY, SKY, SCOWL

          this.debug.log(
            'RULE: S* %s%s -> S* %s%s',
            tables.signInputTable1[p],
            tables.signInputTable2[p],
            tables.signInputTable1[p-12],
            tables.signInputTable2[p-12]
          );
          this.phonemeindex[pos] = p-12;
        } else if ((pf & FLAG_PLOSIVE) === 0) {
          p = this.phonemeindex[pos];
          if (p === 53) {
            // Example: NEW, DEW, SUE, ZOO, THOO, TOO
            rule_alveolar_uw(pos);
          } else if (p === 42) {
            rule_ch(pos);
          } // Example: CHEW
          else if (p === 44) {
            // Example: JAY
            rule_j(pos);
          }
        }

        if (p === 69 || p === 57) { // 'T', 'D'
          // RULE: Soften T following vowel
          // NOTE: This rule fails for cases such as "ODD"
          //       <UNSTRESSED VOWEL> T <PAUSE> -> <UNSTRESSED VOWEL> DX <PAUSE>
          //       <UNSTRESSED VOWEL> D <PAUSE>  -> <UNSTRESSED VOWEL> DX <PAUSE>
          // Example: PARTY, TARDY
          if ((tables.flags[this.phonemeindex[pos-1]] & FLAG_VOWEL) !== 0) {
            p = this.phonemeindex[pos+1];
            if (!p) {
              p = this.phonemeindex[pos+2];
            }
            if ((tables.flags[p] & FLAG_VOWEL) && !this.stress[pos+1]) {
              change(pos,30, 'Soften T or D following vowel or ER and preceding a pause -> DX');
            }
          }
        }
      }
      pos++;
    } // while
  }

  /**
   * Iterates through the phoneme buffer, copying the stress value from
   * the following phoneme under the following circumstance:
   *     1. The current phoneme is voiced, excluding plosives and fricatives
   *     2. The following phoneme is voiced, excluding plosives and fricatives, and
   *     3. The following phoneme is stressed
   *
   *  In those cases, the stress value+1 from the following phoneme is copied.
   *
   * For example, the word LOITER is represented as LOY5TER, with as stress
   * of 5 on the dipthong OY. This routine will copy the stress value of 6 (5+1)
   * to the L that precedes it.
   */
  CopyStress() {
    // loop thought all the phonemes to be output
    let pos = 0; //mem66
    let Y;
    while((Y = this.phonemeindex[pos]) !== END) {
      // if CONSONANT_FLAG set, skip - only vowels get stress
      if ((tables.flags[Y] & 64) !== 0) {
        Y = this.phonemeindex[pos+1];
        // if the following phoneme is the end, or a vowel, skip
        if ((Y !== END) && (tables.flags[Y] & 128) !== 0) {
          // get the stress value at the next position
          Y = this.stress[pos+1];
          if (Y && ((Y & 128) === 0)) {
            // if next phoneme is stressed, and a VOWEL OR ER
            // copy stress from next phoneme to this one
            this.stress[pos] = Y+1;
          }
        }
      }
      ++pos;
    }
  }

  /**
   * change phonemelength depedendent on stress
   */
  SetPhonemeLength() {
    let position = 0;
    while(this.phonemeindex[position] !== 255) {
      let A = this.stress[position];
      if ((A === 0) || ((A&128) !== 0)) {
        this.phonemeLength[position] = tables.phonemeLengthTable[this.phonemeindex[position]];
      } else {
        this.phonemeLength[position] = tables.phonemeStressedLengthTable[this.phonemeindex[position]];
      }
      position++;
    }
  }

  /**
   * Applies various rules that adjust the lengths of phonemes
   *
   * Lengthen <FRICATIVE> or <VOICED> between <VOWEL> and <PUNCTUATION> by 1.5
   * <VOWEL> <RX | LX> <CONSONANT> - decrease <VOWEL> length by 1
   * <VOWEL> <UNVOICED PLOSIVE> - decrease vowel by 1/8th
   * <VOWEL> <UNVOICED CONSONANT> - increase vowel by 1/2 + 1
   * <NASAL> <STOP CONSONANT> - set nasal = 5, consonant = 6
   * <VOICED STOP CONSONANT> {optional silence} <STOP CONSONANT> - shorten both to 1/2 + 1
   * <LIQUID CONSONANT> <DIPTHONG> - decrease by 2
   */
  AdjustLengths() {
    // LENGTHEN VOWELS PRECEDING PUNCTUATION
    //
    // Search for punctuation. If found, back up to the first vowel, then
    // process all phonemes between there and up to (but not including) the punctuation.
    // If any phoneme is found that is a either a fricative or voiced, the duration is
    // increased by (length * 1.5) + 1

    // loop index
    let X = 0;
    let index;

    while((index = this.phonemeindex[X]) !== END) {
      // not punctuation?
      if((tables.flags[index] & FLAG_PUNCT) === 0) {
        ++X;
        continue;
      }

      let loopIndex = X;

      while (--X && ((tables.flags[this.phonemeindex[X]] & FLAG_VOWEL) === 0)) { /* back up while not a vowel */ }
      if (X === 0) {
        break;
      }

      do {
        // test for vowel
        index = this.phonemeindex[X];

        // test for fricative/unvoiced or not voiced
        if(((tables.flags[index] & FLAG_FRICATIVE) === 0) || ((tables.flags[index] & FLAG_VOICED) !== 0)) { //nochmal überprüfen
          let A = this.phonemeLength[X];
          // change phoneme length to (length * 1.5) + 1
          this.drule_pre('Lengthen <FRICATIVE> or <VOICED> between <VOWEL> and <PUNCTUATION> by 1.5', X);
          this.phonemeLength[X] = (A >> 1) + A + 1;
          this.drule_post(X);
        }
      } while (++X !== loopIndex);
      X++;
    }  // while

    // Similar to the above routine, but shorten vowels under some circumstances

    // Loop through all phonemes
    let loopIndex=0;

    while((index = this.phonemeindex[loopIndex]) !== END) {
      let X = loopIndex;

      if ((tables.flags[index] & FLAG_VOWEL) !== 0) {
        index = this.phonemeindex[loopIndex+1];
        if ((tables.flags[index] & FLAG_CONSONANT) === 0) {
          if ((index === 18) || (index === 19)) { // 'RX', 'LX'
            index = this.phonemeindex[loopIndex+2];
            if ((tables.flags[index] & FLAG_CONSONANT) !== 0) {
              this.drule_pre('<VOWEL> <RX | LX> <CONSONANT> - decrease length of vowel by 1', loopIndex);
              this.phonemeLength[loopIndex]--;
              this.drule_post(loopIndex);
            }
          }
        } else { // Got here if not <VOWEL>
          let flag = (index === END) ? 65 : tables.flags[index]; // 65 if end marker

          if ((flag & FLAG_VOICED) === 0) { // Unvoiced
            // *, .*, ?*, ,*, -*, DX, S*, SH, F*, TH, /H, /X, CH, P*, T*, K*, KX
            if((flag & FLAG_PLOSIVE) !== 0) { // unvoiced plosive
              // RULE: <VOWEL> <UNVOICED PLOSIVE>
              // <VOWEL> <P*, T*, K*, KX>
              this.drule_pre('<VOWEL> <UNVOICED PLOSIVE> - decrease vowel by 1/8th',loopIndex);
              this.phonemeLength[loopIndex] -= (this.phonemeLength[loopIndex] >> 3);
              this.drule_post(loopIndex);
            }
          } else {
            this.drule_pre('<VOWEL> <VOICED CONSONANT> - increase vowel by 1/2 + 1',X-1);
            // decrease length
            let A = this.phonemeLength[loopIndex];
            this.phonemeLength[loopIndex] = (A >> 2) + A + 1;     // 5/4*A + 1
            this.drule_post(loopIndex);
          }
        }
      } else if((tables.flags[index] & FLAG_NASAL) !== 0) { // nasal?
        // RULE: <NASAL> <STOP CONSONANT>
        //       Set punctuation length to 6
        //       Set stop consonant length to 5
        index = this.phonemeindex[++X];
        if (index !== END && ((tables.flags[index] & FLAG_STOPCONS) !== 0)) {
          this.drule('<NASAL> <STOP CONSONANT> - set nasal = 5, consonant = 6');
          this.phonemeLength[X]   = 6; // set stop consonant length to 6
          this.phonemeLength[X-1] = 5; // set nasal length to 5
        }
      } else if((tables.flags[index] & FLAG_STOPCONS) !== 0) { // (voiced) stop consonant?
        // RULE: <VOICED STOP CONSONANT> {optional silence} <STOP CONSONANT>
        //       Shorten both to (length/2 + 1)

        while ((index = this.phonemeindex[++X]) === 0) { /* move past silence */ }

        if (index !== END && ((tables.flags[index] & FLAG_STOPCONS) !== 0)) {
          // FIXME, this looks wrong?
          // RULE: <UNVOICED STOP CONSONANT> {optional silence} <STOP CONSONANT>
          this.drule('<UNVOICED STOP CONSONANT> {optional silence} <STOP CONSONANT> - shorten both to 1/2 + 1');
          this.phonemeLength[X]         = (this.phonemeLength[X] >> 1) + 1;
          this.phonemeLength[loopIndex] = (this.phonemeLength[loopIndex] >> 1) + 1;
          X = loopIndex;
        }
      } else if ((tables.flags[index] & FLAG_LIQUIC) !== 0) { // liquic consonant?
        // RULE: <VOICED NON-VOWEL> <DIPTHONG>
        //       Decrease <DIPTHONG> by 2
        index = this.phonemeindex[X-1]; // prior phoneme;

        // FIXME: The debug code here breaks the rule.
        // prior phoneme a stop consonant>
        if((tables.flags[index] & FLAG_STOPCONS) !== 0)
          this.drule_pre('<LIQUID CONSONANT> <DIPTHONG> - decrease by 2', X);

        this.phonemeLength[X] -= 2; // 20ms
        this.drule_post(X);
      }

      ++loopIndex;
    }
  }

  Code41240() {
    let pos=0;

    while(this.phonemeindex[pos] !== END) {
      let index = this.phonemeindex[pos];

      if ((tables.flags[index] & FLAG_STOPCONS) !== 0) {
        if ((tables.flags[index] & FLAG_PLOSIVE) !== 0) {
          let A;
          let X = pos;
          while (!(A = this.phonemeindex[++X])){ /* Skip pause */ }

          if (A !== END) {
            if ((tables.flags[A] & 8) || (A === 36) || (A === 37)) { // '/H' '/X'
              ++pos; continue;
            }
          }
        }
        this.Insert(pos+1, index+1, tables.phonemeLengthTable[index+1], this.stress[pos]);
        this.Insert(pos+2, index+2, tables.phonemeLengthTable[index+2], this.stress[pos]);
        pos += 2;
      }
      ++pos;
    }
  }

  InsertBreath() {
    let mem54 = new UInt8(255);
    let len = new UInt8();
    let index = new UInt8(); //variable Y

    let pos = new UInt8();

    while(index.set(this.phonemeindex[pos.get()]).get() !== END) {
      len.inc(this.phonemeLength[pos.get()]);
      if (len.get() < 232) {
        if (index.get() === BREAK) {
        } else if ((tables.flags[index.get()] & FLAG_PUNCT) === 0) {
          if (index.get() === 0) {
            mem54.set(pos.get());
          }
        } else {
          len.set(0);
          this.Insert(pos.inc().get(), BREAK, 0, 0);
        }
      } else {
        pos.set(mem54.get());
        this.phonemeindex[pos.get()]  = 31;   // 'Q*' glottal stop
        this.phonemeLength[pos.get()] = 4;
        this.stress[pos.get()] = 0;

        len.set(0);
        this.Insert(pos.inc().get(), BREAK, 0, 0);
      }
      pos.inc();
    }
  }

  PrepareOutput() {
    let srcpos  = 0; // Position in source
    let destpos = 0; // Position in output

    while(1) {
      let A = this.phonemeindex[srcpos];
      this.phonemeIndexOutput[destpos] = A;
      switch(A) {
        case END:
          this.Render();
          return;
        case BREAK:
          this.phonemeIndexOutput[destpos] = END;
          this.Render();
          destpos = 0;
          break;
        case 0:
          break;
        default:
          this.phonemeLengthOutput[destpos] = this.phonemeLength[srcpos];
          this.stressOutput[destpos]        = this.stress[srcpos];
          ++destpos;
      }
      ++srcpos;
    }
  }

  /** ASSIGN PITCH CONTOUR
   *
   * This subtracts the F1 frequency from the pitch to create a
   * pitch contour. Without this, the output would be at a single
   * pitch level (monotone).
   */
  AssignPitchContour () {
    for(let i = 0; i < 256; i++) {
      // subtract half the frequency of the formant 1.
      // this adds variety to the voice
      this.pitches[i] -= (this.frequency1[i] >> 1);
    }
  }

  /**
   * RESCALE AMPLITUDE
   *
   * Rescale volume from a linear scale to decibels.
   */
  RescaleAmplitude () {
    const amplitudeRescale = [
      0x00, 0x01, 0x02, 0x02, 0x02, 0x03, 0x03, 0x04,
      0x04, 0x05, 0x06, 0x08, 0x09, 0x0B, 0x0D, 0x0F,
      0x00  //17 elements?
    ];
    for(let i = 255; i >= 0; i--) {
      this.amplitude1[i] = amplitudeRescale[this.amplitude1[i]];
      this.amplitude2[i] = amplitudeRescale[this.amplitude2[i]];
      this.amplitude3[i] = amplitudeRescale[this.amplitude3[i]];
    }
  }

  /**
   * RENDER THE PHONEMES IN THE LIST
   *
   * The phoneme list is converted into sound through the steps:
   *
   * 1. Copy each phoneme <length> number of times into the frames list,
   *    where each frame represents 10 milliseconds of sound.
   *
   * 2. Determine the transitions lengths between phonemes, and linearly
   *    interpolate the values across the frames.
   *
   * 3. Offset the pitches by the fundamental frequency.
   *
   * 4. Render the each frame.
   */
  Render () {
    if (this.phonemeIndexOutput[0] === 255) {
      return; //exit if no data
    }

    this.CreateFrames();
    let t = this.CreateTransitions();

    if (!this.singmode) {
      this.AssignPitchContour();
    }
    this.RescaleAmplitude();

    this.PrintOutput(
      this.sampledConsonantFlag,
      this.frequency1,
      this.frequency2,
      this.frequency3,
      this.amplitude1,
      this.amplitude2,
      this.amplitude3,
      this.pitches
    );

    this.ProcessFrames(t);
  }

  /** CREATE FRAMES
   *
   * The length parameter in the list corresponds to the number of frames
   * to expand the phoneme to. Each frame represents 10 milliseconds of time.
   * So a phoneme with a length of 7 = 7 frames = 70 milliseconds duration.
   *
   * The parameters are copied from the phoneme to the frame verbatim.
   */
  CreateFrames () {
    /**
     * Create a rising or falling inflection 30 frames prior to index X.
     * A rising inflection is used for questions, and a falling inflection is used for statements.
     */
    const AddInflection = (inflection, pos) => {
      let A = new UInt8(0);
      // store the location of the punctuation
      let end = pos;
      if (pos < 30) {
        pos = 0;
      } else {
        pos -= 30;
      }
      // FIXME: Explain this fix better, it's not obvious
      // ML : A =, fixes a problem with invalid pitch with '.'
      while (A.set(this.pitches[pos]).get() === 127) {
        ++pos;
      }

      while (pos !== end) {
        // add the inflection direction
        A.inc(inflection);

        // set the inflection
        this.pitches[pos] = A.get();

        while ((++pos !== end) && this.pitches[pos] === 255) { /* keep looping */}
      }
    };

    let phase1 = 0;
    let X = new UInt8(0);
    let i = 0;
    while(i < 256) {
      // get the phoneme at the index
      let phoneme = this.phonemeIndexOutput[i];

      // if terminal phoneme, exit the loop
      if (phoneme === 255) break;

      if (phoneme === PHONEME_PERIOD) {
        AddInflection(RISING_INFLECTION, X.get());
      } else if (phoneme === PHONEME_QUESTION) {
        AddInflection(FALLING_INFLECTION, X.get());
      }

      // get the stress amount (more stress = higher pitch)
      phase1 = tables.tab47492[this.stressOutput[i] + 1];

      // get number of frames to write
      let phase2 = this.phonemeLengthOutput[i];

      // copy from the source to the frames list
      do {
        this.frequency1[X.get()] = this.freq1data[phoneme];     // F1 frequency
        this.frequency2[X.get()] = this.freq2data[phoneme];     // F2 frequency
        this.frequency3[X.get()] = tables.freq3data[phoneme];     // F3 frequency
        this.amplitude1[X.get()] = tables.ampl1data[phoneme];     // F1 amplitude
        this.amplitude2[X.get()] = tables.ampl2data[phoneme];     // F2 amplitude
        this.amplitude3[X.get()] = tables.ampl3data[phoneme];     // F3 amplitude
        this.sampledConsonantFlag[X.get()] = tables.sampledConsonantFlags[phoneme]; // phoneme data for sampled consonants
        this.pitches[X.get()] = this.pitch + phase1;      // pitch
        X.inc();
      } while(--phase2 !== 0);
      ++i;
    }
  }

  /**
   * CREATE TRANSITIONS.
   *
   * Linear transitions are now created to smoothly connect each
   * phoeneme. This transition is spread between the ending frames
   * of the old phoneme (outBlendLength), and the beginning frames
   * of the new phoneme (inBlendLength).
   *
   * To determine how many frames to use, the two phonemes are
   * compared using the blendRank[] table. The phoneme with the
   * smaller score is used. In case of a tie, a blend of each is used:
   *
   *      if blendRank[phoneme1] ==  blendRank[phomneme2]
   *          // use lengths from each phoneme
   *          outBlendFrames = outBlend[phoneme1]
   *          inBlendFrames = outBlend[phoneme2]
   *      else if blendRank[phoneme1] < blendRank[phoneme2]
   *          // use lengths from first phoneme
   *          outBlendFrames = outBlendLength[phoneme1]
   *          inBlendFrames = inBlendLength[phoneme1]
   *      else
   *          // use lengths from the second phoneme
   *          // note that in and out are swapped around!
   *          outBlendFrames = inBlendLength[phoneme2]
   *          inBlendFrames = outBlendLength[phoneme2]
   *
   *  Blend lengths can't be less than zero.
   *
   * For most of the parameters, SAM interpolates over the range of the last
   * outBlendFrames-1 and the first inBlendFrames.
   *
   * The exception to this is the Pitch[] parameter, which is interpolates the
   * pitch from the center of the current phoneme to the center of the next
   * phoneme.
   */
  CreateTransitions() {
    //written by me because of different table positions.
    // mem[47] = ...
    // 168=pitches
    // 169=frequency1
    // 170=frequency2
    // 171=frequency3
    // 172=amplitude1
    // 173=amplitude2
    // 174=amplitude3
    const Read = (p, Y) => {
      switch(p) {
        case 168: return this.pitches[Y];
        case 169: return this.frequency1[Y];
        case 170: return this.frequency2[Y];
        case 171: return this.frequency3[Y];
        case 172: return this.amplitude1[Y];
        case 173: return this.amplitude2[Y];
        case 174: return this.amplitude3[Y];
      }
      this.debug.error('Error reading from tables');

      return 0;
    };

    const Write = (p, Y, value) => {
      switch (p) {
        case 168: this.pitches[Y] = value; return;
        case 169: this.frequency1[Y] = value; return;
        case 170: this.frequency2[Y] = value; return;
        case 171: this.frequency3[Y] = value; return;
        case 172: this.amplitude1[Y] = value; return;
        case 173: this.amplitude2[Y] = value; return;
        case 174: this.amplitude3[Y] = value; return;
      }
      this.debug.error('Error writing to tables');
    };

    // linearly interpolate values
    const interpolate = (width, table, frame, mem53) => {
      let sign      = (mem53 < 0);
      let remainder = new UInt8(Math.abs(mem53) % width);
      let div       = new UInt8(Math.floor(mem53 / width));

      let error = new UInt8(0);
      let pos   = width;
      let val   = new UInt8(Read(table, frame) + div.get());

      while(--pos) {
        error.inc(remainder.get());
        if (error.get() >= width) {
          // accumulated a whole integer error, so adjust output
          error.dec(width);
          if (sign) {
            val.dec();
          } else if (val.get()) {
            // if input is 0, we always leave it alone
            val.inc();
          }
        }
        Write(table, ++frame, val.get()); // Write updated value back to next frame.
        val.inc(div.get());
      }
    };

    const interpolate_pitch = (width, pos, mem49, phase3) => {
      // unlike the other values, the pitches[] interpolates from
      // the middle of the current phoneme to the middle of the
      // next phoneme

      // half the width of the current and next phoneme
      let cur_width  = Math.floor(this.phonemeLengthOutput[pos] / 2);
      let next_width = Math.floor(this.phonemeLengthOutput[pos+1] / 2);
      // sum the values
      width = cur_width + next_width;
      let pitch = this.pitches[next_width + mem49] - this.pitches[mem49 - cur_width];
      interpolate(width, 168, phase3, pitch);
    };

    let phase1;
    let phase2;
    let mem49 = new UInt8(0);
    let pos = new UInt8(0);
    while(1) {
      let phoneme      = this.phonemeIndexOutput[pos.get()];
      let next_phoneme = this.phonemeIndexOutput[pos.get()+1];

      if (next_phoneme === 255) {
        break; // 255 == end_token
      }

      // get the ranking of each phoneme
      let next_rank = tables.blendRank[next_phoneme];
      let rank      = tables.blendRank[phoneme];

      // compare the rank - lower rank value is stronger
      if (rank === next_rank) {
        // same rank, so use out blend lengths from each phoneme
        phase1 = tables.outBlendLength[phoneme];
        phase2 = tables.outBlendLength[next_phoneme];
      } else if (rank < next_rank) {
        // next phoneme is stronger, so us its blend lengths
        phase1 = tables.inBlendLength[next_phoneme];
        phase2 = tables.outBlendLength[next_phoneme];
      } else {
        // current phoneme is stronger, so use its blend lengths
        // note the out/in are swapped
        phase1 = tables.outBlendLength[phoneme];
        phase2 = tables.inBlendLength[phoneme];
      }

      mem49.inc(this.phonemeLengthOutput[pos.get()]);

      let speedcounter = new UInt8(mem49.get() + phase2);
      let phase3       = new UInt8(mem49.get() - phase1);
      let transition   = new UInt8(phase1 + phase2); // total transition?

      if (((transition.get() - 2) & 128) === 0) {
        interpolate_pitch(transition.get(), pos.get(), mem49.get(), phase3.get());
        let table = 169;
        while (table < 175) {
          // tables:
          // 168  pitches[]
          // 169  frequency1
          // 170  frequency2
          // 171  frequency3
          // 172  amplitude1
          // 173  amplitude2
          // 174  amplitude3

          let value = Read(table, speedcounter.get()) - Read(table, phase3.get());
          interpolate(transition.get(), table, phase3.get(), value);
          table++;
        }
      }
      pos.inc();
    }

    // add the length of this phoneme
    return (mem49.get() + this.phonemeLengthOutput[pos.get()]) & 0xFF;
  }

  /**
   * PROCESS THE FRAMES
   *
   * In traditional vocal synthesis, the glottal pulse drives filters, which
   * are attenuated to the frequencies of the formants.
   *
   * SAM generates these formants directly with sin and rectangular waves.
   * To simulate them being driven by the glottal pulse, the waveforms are
   * reset at the beginning of each glottal pulse.
   */
  ProcessFrames(mem48) {
    let oldTimeTableIndex = 0;
    const Output = (index, A) => {
      // timetable for more accurate c64 simulation
      const timetable = [
        [162, 167, 167, 127, 128],
        [226, 60, 60, 0, 0],
        [225, 60, 59, 0, 0],
        [200, 0, 0, 54, 55],
        [199, 0, 0, 54, 54]
      ];
      this.bufferpos += timetable[oldTimeTableIndex][index];
      oldTimeTableIndex = index;
      // write a little bit in advance
      for (let k = 0; k < 5; k++) {
        this.buffer[Math.floor(this.bufferpos / 50) + k] = (A & 15) * 16;
      }
    };
    const CombineGlottalAndFormants = (phase1, phase2, phase3, Y) => {
      let tmp; // unsigned int
      tmp   = tables.multtable[tables.sinus[phase1]     | this.amplitude1[Y]];
      tmp  += tables.multtable[tables.sinus[phase2]     | this.amplitude2[Y]];
      tmp  += tmp > 255 ? 1 : 0; // if addition above overflows, we for some reason add one;
      tmp  += tables.multtable[tables.rectangle[phase3] | this.amplitude3[Y]];
      tmp  += 136;
      tmp >>= 4; // Scale down to 0..15 range of C64 audio.

      Output(0, tmp & 0xf);
    };

    const RenderSample = (mem66, consonantFlag, mem49) => {
      const RenderVoicedSample = (hi, off, phase1) => {
        hi = hi & 0xFFFF; // unsigned short
        off = off & 0xFF; // unsigned char
        phase1 = phase1 & 0xFF; // unsigned char
        do {
          let sample = tables.sampleTable[hi+off];
          let bit = 8;
          do {
            if ((sample & 128) !== 0) {
              Output(3, 26);
            } else {
              Output(4, 6);
            }
            sample <<= 1;
          } while(--bit !== 0);
          off++;
        } while (((++phase1) & 0xFF) !== 0);

        return off;
      };

      const RenderUnvoicedSample = (hi, off, mem53) => {
        hi = hi & 0xFFFF; // unsigned short
        off = off & 0xFF; // unsigned char
        mem53 = mem53 & 0xFF; // unsigned char
        do {
          let bit = 8;
          let sample = tables.sampleTable[hi+off];
          do {
            if ((sample & 128) !== 0) {
              Output(2, 5);
            }
            else {
              Output(1, mem53);
            }
            sample <<= 1;
          } while (--bit !== 0);
        } while (((++off) & 0xFF) !== 0);
      };

      // mem49 == current phoneme's index - unsigned char

      // mask low three bits and subtract 1 get value to
      // convert 0 bits on unvoiced samples.
      let hibyte = (consonantFlag & 7) - 1;

      // determine which offset to use from table { 0x18, 0x1A, 0x17, 0x17, 0x17 }
      // T, S, Z                0          0x18
      // CH, J, SH, ZH          1          0x1A
      // P, F*, V, TH, DH       2          0x17
      // /H                     3          0x17
      // /X                     4          0x17

      let hi = hibyte * 256; // unsigned short
      // voiced sample?
      let pitch = consonantFlag & 248; // unsigned char
      if(pitch === 0) {
        // voiced phoneme: Z*, ZH, V*, DH
        pitch = this.pitches[mem49 & 0xFF] >> 4;
        return RenderVoicedSample(hi, mem66, pitch ^ 255);
      }
      RenderUnvoicedSample(hi, pitch ^ 255, tables.tab48426[hibyte]);
      return mem66;
    };

    let speedcounter = new UInt8(72);
    let phase1 = new UInt8();
    let phase2 = new UInt8();
    let phase3 = new UInt8();
    let mem66 = new UInt8();
    let Y = new UInt8();
    let glottal_pulse = new UInt8(this.pitches[0]);
    let mem38 = new UInt8(glottal_pulse.get() - (glottal_pulse.get() >> 2)); // mem44 * 0.75

    while(mem48) {
      let flags = this.sampledConsonantFlag[Y.get()];

      // unvoiced sampled phoneme?
      if ((flags & 248) !== 0) {
        mem66.set(RenderSample(mem66.get(), flags, Y.get()));
        // skip ahead two in the phoneme buffer
        Y.inc(2);
        mem48 -= 2;
        speedcounter.set(this.getOptions().speed);
      } else {
        CombineGlottalAndFormants(phase1.get(), phase2.get(), phase3.get(), Y.get());

        speedcounter.dec();
        if (speedcounter.get() === 0) {
          Y.inc(); //go to next amplitude
          // decrement the frame count
          mem48--;
          if(mem48 === 0) {
            return;
          }
          speedcounter.set(this.getOptions().speed);
        }

        glottal_pulse.dec();

        if(glottal_pulse.get() !== 0) {
          // not finished with a glottal pulse

          mem38.dec();
          // within the first 75% of the glottal pulse?
          // is the count non-zero and the sampled flag is zero?
          if((mem38.get() !== 0) || (flags === 0)) {
            // reset the phase of the formants to match the pulse
            phase1.inc(this.frequency1[Y.get()]);
            phase2.inc(this.frequency2[Y.get()]);
            phase3.inc(this.frequency3[Y.get()]);
            continue;
          }

          // voiced sampled phonemes interleave the sample with the
          // glottal pulse. The sample flag is non-zero, so render
          // the sample for the phoneme.
          mem66.set(RenderSample(mem66.get(), flags, Y.get()));
        }
      }

      glottal_pulse.set(this.pitches[Y.get()]);
      mem38.set(glottal_pulse.get() - (glottal_pulse.get() >> 2)); // mem44 * 0.75

      // reset the formant wave generators to keep them in
      // sync with the glottal pulse
      phase1.set(0);
      phase2.set(0);
      phase3.set(0);
    }
  }

  PrintPhonemes (phonemeindex, phonemeLength, stress) {
    let i = 0;
    this.debug.log('===========================================');
    this.debug.log('Internal Phoneme presentation:');
    this.debug.log(' idx    phoneme  length  stress');
    this.debug.log('------------------------------');

    while((phonemeindex[i] !== 255) && (i < 255))
    {
      if (phonemeindex[i] < 81)
      {
        this.debug.log(' %3i      %s%s      %3i       %i',
          phonemeindex[i],
          String.fromCharCode(tables.signInputTable1[phonemeindex[i]]),
          String.fromCharCode(tables.signInputTable2[phonemeindex[i]]),
          phonemeLength[i],
          stress[i]
        );
      } else
      {
        this.debug.log(' %3i      ??      %3i       %i', phonemeindex[i], phonemeLength[i], stress[i]);
      }
      i++;
    }
    this.debug.log("===========================================");
  }

  PrintOutput(flag, f1, f2, f3, a1, a2, a3, p) {
    this.debug.log('===========================================');
    this.debug.log('Final data for speech output:');
    let i = 0;
    this.debug.log(' flags ampl1 freq1 ampl2 freq2 ampl3 freq3 pitch');
    this.debug.log('------------------------------------------------');
    while(i < 255)
    {
      this.debug.log('%5i %5i %5i %5i %5i %5i %5i %5i', flag[i], a1[i], f1[i], a2[i], f2[i], a3[i], f3[i], p[i]);
      i++;
    }
    this.debug.log('===========================================');
  }

  drule(str) {
    this.debug.log('RULE: %s', str);
  }

  drule_pre(descr, X) {
    this.drule(descr);
    this.debug.log('PRE phoneme %d (%s) length %d',
      X,
      String.fromCharCode(tables.signInputTable1[this.phonemeindex[X]], tables.signInputTable2[this.phonemeindex[X]]),
      this.phonemeLength[X]
    );
  }

  drule_post(X) {
    this.debug.log('POST phoneme %d (%s) length %d',
      X,
      String.fromCharCode(tables.signInputTable1[this.phonemeindex[X]], tables.signInputTable2[this.phonemeindex[X]]),
      this.phonemeLength[X]
    );
  }
}
